#include <stdio.h>

/* #include <stdlib.h> */

#include "config_xor.h"

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_MEMORY_H
#include <memory.h>
#endif

#if defined(HAVE_MLOCK) && !defined(HAVE_BROKEN_MLOCK)
#include <sys/mman.h>
#endif

#include "sh_tiger.h"

#include "sh_unix.h"
#include "sh_error.h"

typedef unsigned long sh_word32;
typedef unsigned char sh_byte;

void
tiger(sh_word32 *str, sh_word32 length, sh_word32 res[6]);


/* this is the wrapper function -- not part of the tiger reference
 * implementation
 */
char * sh_tiger_hash (char * filename, TigerType what, unsigned long Length)
{
  int  fd;
  int  i, j;
  int  count;
  static char out[51];
  sh_byte buffer[PRIV_MAX+2];
  sh_byte bbuf[64];

  unsigned long ncount, nblocks = 0;
  unsigned long  t, msb, lsb;

  sh_word32 res[6];


  /* initialize 
   */
  res[0]=0x89ABCDEF;
  res[1]=0x01234567;
  res[2]=0x76543210;
  res[3]=0xFEDCBA98;
  res[4]=0xC3B2E187;
  res[5]=0xF096A5B4;

  if (what == TIGER_FILE) {
    fd = sh_unix_fileopen (filename, 'r', 
                           SH_ISDATA, ShDFLevel[SH_ERR_T_FILE]);
    if (fd == (-1)) 
      return _("0000000000000000 000000000000000 000000000000000");

#if defined(HAVE_MLOCK) && !defined(HAVE_BROKEN_MLOCK)
    if (skey->mlock_failed == GOOD) 
      {
        if ( (-1) == mlock( buffer, (PRIV_MAX+2) * sizeof(sh_byte)) ) 
          skey->mlock_failed = BAD;
      }
#else
    if (skey->mlock_failed == GOOD) {
      skey->mlock_failed = BAD;
    }
#endif


    while (1) {

      count = sh_unix_fileread (fd, buffer, PRIV_MAX);

      nblocks += count / 64; /* number of 64-byte words */
      ncount  += count;      /* number of bytes         */

      if (count < PRIV_MAX) 
	break;

      tiger((sh_word32*) buffer, ( count < PRIV_MAX ? count : PRIV_MAX), res);

    }

    /* copy incomplete block
     */
    j = 0;
    for (i = (count/64)*64; i < count; ++i)
      {
	bbuf[j] = buffer[i];
	++j;
      }

    for (i = count; i < PRIV_MAX; ++i) buffer[i] = '\0';

    tiger((sh_word32*) buffer, (count/64)*64, res);

    /* code fragment from gnupg
     */
    msb = 0;
    t = nblocks;
    if( (lsb = t << 6) < t ) /* multiply by 64 to make a byte count */
      msb++;
    msb += t >> 26;
    t = lsb;
    if( (lsb = t + count) < t ) /* add the count */
      msb++;
    t = lsb;
    if( (lsb = t << 3) < t ) /* multiply by 8 to make a bit count */
      msb++;
    msb += t >> 29;

    if( j < 56 ) { /* enough room */
        bbuf[j++] = 0x01; /* pad */
        while( j < 56 )
            bbuf[j++] = 0;  /* pad */
    }
    else { /* need one extra block */
        bbuf[j++] = 0x01; /* pad character */
        while( j < 64 )
            bbuf[j++] = 0;
	tiger((sh_word32*) bbuf, 64, res);
        memset(bbuf, 0, 56 ); /* fill next block with zeroes */
    }

    bbuf[56] = lsb      ;
    bbuf[57] = lsb >>  8;
    bbuf[58] = lsb >> 16;
    bbuf[59] = lsb >> 24;
    bbuf[60] = msb      ;
    bbuf[61] = msb >>  8;
    bbuf[62] = msb >> 16;
    bbuf[63] = msb >> 24;

    tiger((sh_word32*) bbuf, 64, res);

    memset (bbuf,   0, 64);
    memset (buffer, 0, PRIV_MAX);
#if defined(HAVE_MLOCK) && !defined(HAVE_BROKEN_MLOCK)
    munlock( buffer, (PRIV_MAX+2) * sizeof(sh_byte));
#endif
    sh_unix_fileclose (fd);

    sprintf(out,                                     /* known to fit  */
            _("%08lX%08lX %08lX%08lX %08lX%08lX"),
            (sh_word32)(res[1]), 
            (sh_word32)(res[0]), 
            (sh_word32)(res[3]), 
            (sh_word32)(res[2]), 
            (sh_word32)(res[5]), 
            (sh_word32)(res[4]) );
    out[50] = '\0';
    return (out);
 
  }


  if (what == TIGER_DATA) 
    {

      tiger((sh_word32*) filename,  (sh_word32) Length, res); 

      sprintf(out, _("%08lX%08lX %08lX%08lX %08lX%08lX"), 
	      (sh_word32)(res[1]), 
	      (sh_word32)(res[0]), 
	      (sh_word32)(res[3]), 
	      (sh_word32)(res[2]), 
	      (sh_word32)(res[5]), 
	      (sh_word32)(res[4]) );
      return (out);
    }

  return _("0000000000000000 000000000000000 000000000000000");

}

